/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.dialog;

import static org.eclipse.jface.dialogs.IDialogConstants.CANCEL_ID;
import static org.eclipse.jface.dialogs.IDialogConstants.CANCEL_LABEL;
import static org.eclipse.jface.dialogs.IDialogConstants.OK_ID;
import static org.eclipse.jface.dialogs.IDialogConstants.OK_LABEL;
import static org.eclipse.wb.swt.ResourceManager.decorateImage;
import static org.eclipse.wb.swt.ResourceManager.getPluginImage;
import static org.eclipse.wb.swt.SWTResourceManager.BOTTOM_RIGHT;

import java.util.ArrayList;
import java.util.List;

import org.conqat.ide.commons.ui.jface.TreeContentProviderBase;
import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TreePath;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;

/**
 * This dialog offers a common dialog to show multi-root tree structures and
 * allows user to select multiple items in the tree. It also offers an interface
 * to edit the tree on the fly.
 * 
 * @author mou
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 18515CD76B22C0333B81F1D147F24A58
 */
public class ElementTreeMultiSelectDialog extends TitleAreaDialog {

	/** Default content provider */
	private static class EmptyTreeContentProvider extends TreeContentProviderBase {

		/** {@inheritDoc} */
		@Override
		public Object[] getChildren(Object parentElement) {
			return new Object[0];
		}
	}

	/** Label provider which adds selected icon on the original icon */
	private class DecorateLabelProvider extends LabelProvider {

		/** {@inheritDoc} */
		@Override
		public String getText(Object element) {
			return labelSupport.getText(element);
		}

		/** {@inheritDoc} */
		@Override
		public Image getImage(Object element) {
			Image image = labelSupport.getImage(element);
			if(!selectedElements.contains(element))
				return image;

			Image icon = getPluginImage("org.eclipse.ui", "/icons/full/ovr16/pinned_ovr.gif");
			if(icon == null)
				return image;

			return decorateImage(image, icon, BOTTOM_RIGHT);
		}
	}

	/** roots of the tree */
	private ArrayList<Object> rootElements;

	/** selected element */
	private ArrayList<Object> selectedElements;

	/** content provider for the roots */
	private ITreeContentProvider contentSupport;

	/** label provider for the elements */
	private ILabelProvider labelSupport;

	/** The gui composite */
	private ElementTreeMultiSelectDialogGUI gui;

	/** Constructor */
	public ElementTreeMultiSelectDialog(Shell parentShell, List<?> roots, List<?> selected,
			ITreeContentProvider contentProvider, ILabelProvider labelProvider) {
		super(parentShell);

		if(roots == null)
			rootElements = new ArrayList<Object>();
		else
			rootElements = new ArrayList<Object>(roots);

		if(selected == null)
			selectedElements = new ArrayList<Object>();
		else
			selectedElements = new ArrayList<Object>(selected);

		contentSupport = contentProvider;
		labelSupport = labelProvider;
	}

	/** {@inheritDoc} */
	@Override
	protected Control createDialogArea(Composite parent) {
		setHelpAvailable(false);

		setTitle("Select");
		setMessage("Select one or more elements");

		// create controls
		Composite area = (Composite)super.createDialogArea(parent);
		gui = new ElementTreeMultiSelectDialogGUI(area, SWT.NONE);

		gui.getSelectButton().addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				selectElement();
			}
		});

		gui.getDeselectButton().addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				deselectElement();
			}
		});

		gui.getElementTreeViewer().addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				selectElement();
			}
		});

		gui.getSelectTreeViewer().addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				deselectElement();
			}
		});

		// label, content and input
		gui.getElementTreeViewer().setLabelProvider(new DecorateLabelProvider());
		gui.getSelectTreeViewer().setLabelProvider(new DecorateLabelProvider());

		gui.getElementTreeViewer().setContentProvider(
				new MultiRootTreeContentProvider(rootElements, contentSupport));
		gui.getSelectTreeViewer().setContentProvider(
				new MultiRootTreeContentProvider(selectedElements, new EmptyTreeContentProvider()));

		gui.getElementTreeViewer().setInput(rootElements);
		gui.getSelectTreeViewer().setInput(selectedElements);

		return area;
	}

	/** {@inheritDoc} */
	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		createButton(parent, OK_ID, OK_LABEL, false);
		createButton(parent, CANCEL_ID, CANCEL_LABEL, true);
	}

	/** {@inheritDoc} */
	@Override
	protected Point getInitialSize() {
		return new Point(640, 480);
	}

	/** Returns rootElements. */
	public List<Object> getRootElements() {
		return rootElements;
	}

	/** Returns selectedElements. */
	public List<Object> getSelectedElements() {
		return selectedElements;
	}

	/** Returns contentSupport. */
	public IStructuredContentProvider getContentProvider() {
		return contentSupport;
	}

	/** Returns labelSupport. */
	public ILabelProvider getLabelProvider() {
		return labelSupport;
	}

	/** Add selected elements to the select list. */
	private void selectElement() {
		try {
			// find the path to the element
			ITreeSelection selection = (ITreeSelection)gui.getElementTreeViewer().getSelection();
			if(selection.getPaths() == null)
				return;

			for(TreePath path : selection.getPaths()) {
				if(selectedElements.contains(path.getLastSegment()))
					continue;

				selectedElements.add(path.getLastSegment());
				gui.getSelectTreeViewer().add(selectedElements, path.getLastSegment());
				gui.getElementTreeViewer().update(path.getLastSegment(), null);
			}

		} catch(Exception e2) {
			e2.printStackTrace();
		}
	}

	/** Remove selected elements from the select list */
	private void deselectElement() {
		try {
			// find the path to the element
			ITreeSelection selection = (ITreeSelection)gui.getSelectTreeViewer().getSelection();
			if(selection.getPaths() == null)
				return;

			for(TreePath path : selection.getPaths()) {
				if(selectedElements.remove(path.getLastSegment())) {
					gui.getSelectTreeViewer().remove(path.getLastSegment());
					gui.getElementTreeViewer().update(path.getLastSegment(), null);
				}
			}

		} catch(Exception e2) {
			e2.printStackTrace();
		}
	}
}
