/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editpart;

import static org.eclipse.draw2d.ColorConstants.red;
import static org.eclipse.draw2d.Graphics.LINE_SOLID;
import static org.eclipse.gef.EditPolicy.COMPONENT_ROLE;
import static org.eclipse.gef.EditPolicy.CONNECTION_BENDPOINTS_ROLE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.ui.editpart.ExtendedLayerRootEditPart.DECORATION_LAYER;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getConnectionPoints;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.draw2d.AbsoluteBendpoint;
import org.eclipse.draw2d.BendpointConnectionRouter;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.MidpointLocator;
import org.eclipse.draw2d.PolygonDecoration;
import org.eclipse.draw2d.PolylineConnection;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.editparts.AbstractConnectionEditPart;
import org.eclipse.gef.editpolicies.BendpointEditPolicy;
import org.eclipse.gef.requests.BendpointRequest;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.fortiss.tooling.base.model.base.ConnectionSegmentBase;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.ui.editpart.command.bendpoint.CreateCommand;
import org.fortiss.tooling.base.ui.editpart.command.bendpoint.DeleteCommand;
import org.fortiss.tooling.base.ui.editpart.command.bendpoint.MoveCommand;
import org.fortiss.tooling.base.ui.editpart.policy.RemoveEditPolicy;
import org.fortiss.tooling.base.ui.layout.IDiagramLayoutConfiguration;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation.ESeverity;
import org.fortiss.tooling.kernel.ui.extension.IModelElementHandler;
import org.fortiss.tooling.kernel.ui.service.IMarkerService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

/**
 * Base class for connection edit parts. It provides the figure, labels for the
 * connection name and a highlighted simulation value, as well as, bend point
 * handling.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: FF7213723C6970DB4AD073CC7161E110
 */
public abstract class ConnectionEditPartBase<T extends ConnectionSegmentBase> extends
		AbstractConnectionEditPart {

	/**
	 * Stores the diagram layout configuration. Since connections are contained
	 * in the root layer, the configuration must be stored here. The parent
	 * cannot be used.
	 */
	protected final IDiagramLayoutConfiguration layoutConfiguration;

	/** The underlying model object. */
	protected final T modelElement;

	/** The model element handler to be used for the {@link #modelElement}. */
	protected final IModelElementHandler<T> modelElementHandler;

	/** The connection drawn. */
	protected PolylineConnection connection;

	/** The label drawn next to the connection. */
	protected final Label relationshipLabel;

	/** The simulation value label drawn in highlighted color. */
	protected final Label simulationLabel;

	/** The decoration figure. */
	protected Label decorationFigure;

	/**
	 * The adapter used to receive change notifications for the {@link #modelElement}.
	 */
	// protected final Adapter fullRefreshAdapter = new AdapterImpl() {
	protected final Adapter fullRefreshAdapter = new EContentAdapter() {
		/** {@inheritDoc} */
		@Override
		public void notifyChanged(Notification notification) {
			super.notifyChanged(notification);
			// do not care what happens, just refresh everything
			refresh();
		}
	};

	/** Constructor. */
	@SuppressWarnings("unchecked")
	protected ConnectionEditPartBase(T modelElement, IDiagramLayoutConfiguration layoutConfiguration) {
		this.modelElement = modelElement;
		this.layoutConfiguration = layoutConfiguration;
		setModel(modelElement);
		this.modelElementHandler =
				(IModelElementHandler<T>)IModelElementHandlerService.INSTANCE
						.getModelElementHandler(modelElement);
		this.relationshipLabel = new Label(determineRelationshipLabelFigureText()) {
			/** {@inheritDoc} */
			@Override
			public void setBounds(Rectangle rect) {
				super.setBounds(rect);
				if(decorationFigure != null && decorationFigure.getParent() != null) {
					refreshDecorationFigure();
				}
			}
		};
		this.simulationLabel = new Label("");
		simulationLabel.setForegroundColor(getSimulationLabelColor());
	}

	/** {@inheritDoc} */
	@Override
	protected void createEditPolicies() {
		installEditPolicy(COMPONENT_ROLE, new RemoveEditPolicy());
		installEditPolicy(CONNECTION_BENDPOINTS_ROLE, new ConnectionBendpointEditPolicy());
		// TODO (FH): https://af3.fortiss.org/issues/390
		// implement reconnect
		// installEditPolicy(CONNECTION_ENDPOINTS_ROLE,
		// new ConnectionEndpointEditPolicy());
	}

	/** {@inheritDoc} */
	@Override
	protected PolylineConnection createFigure() {
		connection = new PolylineConnection();

		// arrow at target end point
		connection.setTargetDecoration(new PolygonDecoration());
		connection.setLineStyle(LINE_SOLID);

		BendpointConnectionRouter connectionRouter = new BendpointConnectionRouter();
		connection.setConnectionRouter(connectionRouter);
		return connection;
	}

	/**
	 * Returns the position and size of the decoration figure. The computation
	 * may use the bounds of the base figure or the label figure, since both
	 * bounds were determined first.
	 */
	protected Rectangle determineDecorationFigureBounds() {
		Rectangle r = new Rectangle(relationshipLabel.getBounds());
		r.x -= 16;
		r.y -= 16;
		r.width = useSmallDecorationImage() ? 8 : 16;
		r.height = useSmallDecorationImage() ? 8 : 16;
		return r;
	}

	/** Refreshes the decoration figure using {@link IMarkerService} */
	private void refreshDecorationFigure() {
		Rectangle decorationBounds = determineDecorationFigureBounds();
		decorationFigure.setBounds(decorationBounds);
		decorationFigure.getParent().setConstraint(decorationFigure, decorationBounds);

		Image icon = null;
		ESeverity sev = IMarkerService.INSTANCE.getHighestViolationSeverity(getModel());
		switch(sev) {
			case ERROR:
				icon =
						IMarkerService.INSTANCE.getImageFor(ESeverity.ERROR,
								useSmallDecorationImage()).createImage();
				break;
			case WARNING:
				icon =
						IMarkerService.INSTANCE.getImageFor(ESeverity.WARNING,
								useSmallDecorationImage()).createImage();
				break;
			default:
		}
		decorationFigure.setIcon(icon);
		if(icon != null) {
			String message = "";
			Collection<IConstraintViolation<? extends EObject>> violations =
					IMarkerService.INSTANCE.getViolations(getModel());
			for(IConstraintViolation<? extends EObject> viol : violations) {
				if(viol.getSeverity() == sev) {
					message = viol.getExplanation();
					break;
				}
			}
			decorationFigure.setToolTip(new Label(message));
		}
	}

	/**
	 * Returns whether to use small or big decoration image. Default
	 * implementation uses big image.
	 */
	protected boolean useSmallDecorationImage() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public void activate() {
		super.activate();
		getModel().eAdapters().add(fullRefreshAdapter);
	}

	/** {@inheritDoc} */
	@Override
	public void deactivate() {
		getModel().eAdapters().remove(fullRefreshAdapter);
		super.deactivate();
	}

	/** Returns the model element of this edit part. */
	@Override
	public T getModel() {
		return modelElement;
	}

	/** Sets the simulation value label to the given value. */
	public void setSimulationValue(String value) {
		simulationLabel.setText(value);
		refreshVisuals();
	}

	/** Returns the simulation label color. */
	protected Color getSimulationLabelColor() {
		return red;
	}

	/** {@inheritDoc} */
	@Override
	protected void refreshVisuals() {
		connection.setToolTip(new Label(IModelElementHandlerService.INSTANCE
				.getWrappedNameDescription(getModel())));
		refreshBendpoints();
	}

	/** {@inheritDoc} */
	@Override
	public void addNotify() {
		super.addNotify();
		decorationFigure = createDecorationFigure();
		if(decorationFigure != null) {
			getLayer(DECORATION_LAYER).add(decorationFigure);
		}
	}

	/** {@inheritDoc} */
	@SuppressWarnings("rawtypes")
	@Override
	public Object getAdapter(Class key) {
		// EObjectSelectionUtils requires us to adapt to EObject, since
		// graphical edit parts can be provided as selection as well.
		if(key == EObject.class) {
			return getModel();
		}
		return super.getAdapter(key);
	}

	/**
	 * Creates the decoration figure.
	 */
	protected Label createDecorationFigure() {
		return new Label();
	}

	/**
	 * Determines the text of the relationship label figure during a refresh.
	 * The default implementation returns the model element name.
	 */
	protected String determineRelationshipLabelFigureText() {
		return getModel().getName();
	}

	/** Updates the bend points, based on the model. */
	protected void refreshBendpoints() {
		Points pts = getConnectionPoints(getModel());

		List<AbsoluteBendpoint> figureConstraint = new ArrayList<AbsoluteBendpoint>();
		int size = 0;
		if(pts != null) {
			size = pts.getPointsLength();
			for(int i = 0; i < size; i++) {
				org.fortiss.tooling.base.model.layout.Point pt = pts.getPoints(i);
				figureConstraint.add(new AbsoluteBendpoint(pt.getX(), pt.getY()));
			}
		}

		// Update connection label
		if(connection.getChildren().contains(relationshipLabel)) {
			connection.remove(relationshipLabel);
		}
		relationshipLabel.setText(determineRelationshipLabelFigureText());
		MidpointLocator mpl = new MidpointLocator(connection, size / 2) {
			@Override
			protected Point getReferencePoint() {
				Point p = super.getReferencePoint();
				p.y += DEFAULT_GRID_SIZE;
				return p;
			}
		};
		connection.add(relationshipLabel, mpl);

		// Update simulation label
		if(connection.getChildren().contains(simulationLabel)) {
			connection.remove(simulationLabel);
		}
		if(!"".equals(simulationLabel.getText())) {
			mpl = new MidpointLocator(connection, size / 2) {
				@Override
				protected org.eclipse.draw2d.geometry.Point getReferencePoint() {
					org.eclipse.draw2d.geometry.Point p = super.getReferencePoint();
					p.y += 2 * DEFAULT_GRID_SIZE;
					return p;
				}
			};
			connection.add(simulationLabel, mpl);
		}

		getConnectionFigure().setRoutingConstraint(figureConstraint);
	}

	/** Policy for bend points. */
	class ConnectionBendpointEditPolicy extends BendpointEditPolicy {
		/** {@inheritDoc} */
		@Override
		protected Command getCreateBendpointCommand(BendpointRequest request) {
			Point location = request.getLocation();
			getFigure().translateToRelative(location);
			return new CreateCommand(getModel(), request.getIndex(), location,
					getLayoutConfiguration());
		}

		/** {@inheritDoc} */
		@Override
		protected Command getDeleteBendpointCommand(BendpointRequest request) {
			return new DeleteCommand(getModel(), request.getIndex());
		}

		/** {@inheritDoc} */
		@Override
		protected Command getMoveBendpointCommand(BendpointRequest request) {
			Point location = request.getLocation();
			getFigure().translateToRelative(location);

			return new MoveCommand(getModel(), request.getIndex(), location,
					getLayoutConfiguration());
		}

		/** {@inheritDoc} */
		@Override
		protected void showCreateBendpointFeedback(BendpointRequest request) {
			getLayoutConfiguration().performBendPointLayout(request.getLocation());
			super.showCreateBendpointFeedback(request);
		}

		/** {@inheritDoc} */
		@Override
		protected void showMoveBendpointFeedback(BendpointRequest request) {
			getLayoutConfiguration().performBendPointLayout(request.getLocation());
			super.showMoveBendpointFeedback(request);
		}
	}

	/**
	 * Returns the layout configuration for the diagram containing this
	 * connection edit part. This method delegates does not delegate to the
	 * parent edit part, since the parent is the root layer, unfortunately not
	 * the diagram edit part.
	 */
	public IDiagramLayoutConfiguration getLayoutConfiguration() {
		return layoutConfiguration;
	}
}
