/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.xml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;

import javax.xml.transform.TransformerException;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import org.conqat.lib.commons.filesystem.FileSystemUtils;
import org.conqat.lib.commons.xml.XMLUtils;
import org.conqat.ide.commons.ui.CCSMCommonsActivator;

/**
 * Utility code for dealing with XML in eclipse.
 * 
 * @author hummelb
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: E15B5FD26AD8C96B6A93E78A9C9CC8FB
 */
public class EclipseXMLUtils {

	/**
	 * Writes an XML document to an existing file. The XML will be formatted
	 * (indented) in this process.
	 * 
	 * @param doc
	 *            the document being written.
	 * @param file
	 *            the file to write to
	 * @param monitor
	 *            the (optional) progress monitor.
	 * @throws CoreException
	 *             in case of I/O problems. Also thrown if pretty printing fails
	 *             or the encoding is not supported (both unlikely).
	 */
	public static void write(Document doc, IFile file, IProgressMonitor monitor)
			throws CoreException {
		file.setContents(createPrettyXMLStream(doc), 0, monitor);
		file.setCharset(FileSystemUtils.UTF8_ENCODING, monitor);
	}

	/**
	 * Creates a new file and puts the XML document into it. The XML will be
	 * formatted (indented) in this process.
	 * 
	 * @param doc
	 *            the document being written.
	 * @param file
	 *            the file to write to
	 * @param monitor
	 *            the (optional) progress monitor.
	 * @throws CoreException
	 *             in case of I/O problems. Also thrown if pretty printing fails
	 *             or the encoding is not supported (both unlikely).
	 */
	public static void create(Document doc, IFile file, IProgressMonitor monitor)
			throws CoreException {
		file.create(createPrettyXMLStream(doc), 0, monitor);
		file.setCharset(FileSystemUtils.UTF8_ENCODING, monitor);
	}

	/** Formats the given doc and returns a stream containing its serialization. */
	private static InputStream createPrettyXMLStream(Document doc)
			throws CoreException {
		try {
			return new ByteArrayInputStream(XMLUtils.prettyPrint(doc).getBytes(
					FileSystemUtils.UTF8_ENCODING));
		} catch (TransformerException e) {
			throw new CoreException(new Status(IStatus.ERROR,
					CCSMCommonsActivator.PLUGIN_ID,
					"XSLT Transformation failed.", e));
		} catch (UnsupportedEncodingException e) {
			throw new CoreException(new Status(IStatus.ERROR,
					CCSMCommonsActivator.PLUGIN_ID, "UTF-8 not supported.", e));
		}
	}

	/** Reads an XML file and returns the contained document. */
	public static Document read(IFile file) throws CoreException {
		return read(file, null);
	}

	/**
	 * Reads an XML file while validating to an schema and returns the contained
	 * document.
	 */
	public static Document read(IFile file, URL xsdSchema) throws CoreException {
		InputStream contents = null;
		try {
			contents = file.getContents();
			if (xsdSchema != null) {
				return XMLUtils.parse(new InputSource(contents), xsdSchema);
			}
			return XMLUtils.parse(new InputSource(contents));
		} catch (SAXException e) {
			throw new CoreException(new Status(IStatus.ERROR,
					CCSMCommonsActivator.PLUGIN_ID, "Parsing exception: "
							+ e.getMessage(), e));
		} catch (IOException e) {
			throw new CoreException(new Status(IStatus.ERROR,
					CCSMCommonsActivator.PLUGIN_ID, "IO exception: "
							+ e.getMessage(), e));
		} finally {
			FileSystemUtils.close(contents);
		}
	}
}